@@ -65,7 +65,8 @@ st_delete_options = '''[--all] [--leave-segments]
                     [--container-threads <threads>]
                     [--header <header:value>]
                     [--prefix <prefix>]
-                    [<container> [<object>] [...]]
+                    [--versions]
+                    [<container> [<object>] [--version-id <version_id>] [...]]
 '''
 
 st_delete_help = '''
@@ -77,7 +78,8 @@ Positional arguments:
                         for multiple objects.
 
 Optional arguments:
-  -a, --all             Delete all containers and objects.
+  -a, --all             Delete all containers and objects. Implies --versions.
+  --versions            Delete all versions.
   --leave-segments      Do not delete segments of manifest objects.
   -H, --header <header:value>
                         Adds a custom request header to use for deleting
@@ -89,6 +91,8 @@ Optional arguments:
                         Number of threads to use for deleting containers.
                         Default is 10.
   --prefix <prefix>     Only delete objects beginning with <prefix>.
+  --version-id <version-id>
+                        Delete specific version of a versioned object.
 '''.strip("\n")
 
 
@@ -96,9 +100,14 @@ def st_delete(parser, args, output_manager, return_parser=False):
     parser.add_argument(
         '-a', '--all', action='store_true', dest='yes_all',
         default=False, help='Delete all containers and objects.')
+    parser.add_argument('--versions', action='store_true',
+                        help='delete all versions')
     parser.add_argument(
         '-p', '--prefix', dest='prefix',
         help='Only delete items beginning with <prefix>.')
+    parser.add_argument(
+        '--version-id', action='store', default=None,
+        help='Delete a specific version of a versioned object')
     parser.add_argument(
         '-H', '--header', action='append', dest='header',
         default=[],
@@ -123,11 +132,17 @@ def st_delete(parser, args, output_manager, return_parser=False):
 
     (options, args) = parse_args(parser, args)
     args = args[1:]
+    if options['yes_all']:
+        options['versions'] = True
     if (not args and not options['yes_all']) or (args and options['yes_all']):
         output_manager.error('Usage: %s delete %s\n%s',
                              BASENAME, st_delete_options,
                              st_delete_help)
         return
+    if options['versions'] and len(args) >= 2:
+        exit('--versions option not allowed for object deletes')
+    if options['version_id'] and len(args) < 2:
+        exit('--version-id option only allowed for object deletes')
 
     if options['object_threads'] <= 0:
         output_manager.error(
@@ -227,6 +242,7 @@ st_download_options = '''[--all] [--marker <marker>] [--prefix <prefix>]
                       [--object-threads <threads>] [--ignore-checksum]
                       [--container-threads <threads>] [--no-download]
                       [--skip-identical] [--remove-prefix]
+                      [--version-id <version_id>]
                       [--header <header:value>] [--no-shuffle]
                       [<container> [<object>] [...]]
 '''
@@ -271,6 +287,8 @@ Optional arguments:
                         Example: --header "content-type:text/plain"
   --skip-identical      Skip downloading files that are identical on both
                         sides.
+  --version-id <version-id>
+                        Download specific version of a versioned object.
   --ignore-checksum     Turn off checksum validation for downloads.
   --no-shuffle          By default, when downloading a complete account or
                         container, download order is randomised in order to
@@ -332,6 +350,9 @@ def st_download(parser, args, output_manager, return_parser=False):
         '--skip-identical', action='store_true', dest='skip_identical',
         default=False, help='Skip downloading files that are identical on '
         'both sides.')
+    parser.add_argument(
+        '--version-id', action='store', default=None,
+        help='Download a specific version of a versioned object')
     parser.add_argument(
         '--ignore-checksum', action='store_false', dest='checksum',
         default=True, help='Turn off checksum validation for downloads.')
@@ -372,6 +393,8 @@ def st_download(parser, args, output_manager, return_parser=False):
         output_manager.error('Usage: %s download %s\n%s', BASENAME,
                              st_download_options, st_download_help)
         return
+    if options['version_id'] and len(args) < 2:
+        exit('--version-id option only allowed for object downloads')
 
     if options['object_threads'] <= 0:
         output_manager.error(
@@ -479,7 +502,7 @@ def st_download(parser, args, output_manager, return_parser=False):
 
 st_list_options = '''[--long] [--lh] [--totals] [--prefix <prefix>]
                   [--delimiter <delimiter>] [--header <header:value>]
-                  [<container>]
+                  [--versions] [<container>]
 '''
 
 st_list_help = '''
@@ -499,6 +522,8 @@ Optional arguments:
                         Roll up items with the given delimiter. For containers
                         only. See OpenStack Swift API documentation for what
                         this means.
+  -j, --json            Display listing information in json
+  --versions            Display listing information for all versions
   -H, --header <header:value>
                         Adds a custom request header to use for listing.
 '''.strip('\n')
@@ -506,29 +531,29 @@ Optional arguments:
 
 def st_list(parser, args, output_manager, return_parser=False):
 
-    def _print_stats(options, stats, human):
-        total_count = total_bytes = 0
+    def _print_stats(options, stats, human, totals):
         container = stats.get("container", None)
         for item in stats["listing"]:
             item_name = item.get('name')
-            if not options['long'] and not human:
+            if not options['long'] and not human and not options['versions']:
                 output_manager.print_msg(item.get('name', item.get('subdir')))
             else:
                 if not container:    # listing containers
                     item_bytes = item.get('bytes')
                     byte_str = prt_bytes(item_bytes, human)
                     count = item.get('count')
-                    total_count += count
+                    totals['count'] += count
                     try:
                         meta = item.get('meta')
                         utc = gmtime(float(meta.get('x-timestamp')))
                         datestamp = strftime('%Y-%m-%d %H:%M:%S', utc)
                     except TypeError:
                         datestamp = '????-??-?? ??:??:??'
+                    storage_policy = meta.get('x-storage-policy', '???')
                     if not options['totals']:
                         output_manager.print_msg(
-                            "%5s %s %s %s", count, byte_str,
-                            datestamp, item_name)
+                            "%12s %s %s %-15s %s", count, byte_str,
+                            datestamp, storage_policy, item_name)
                 else:    # list container contents
                     subdir = item.get('subdir')
                     content_type = item.get('content_type')
@@ -543,20 +568,17 @@ def st_list(parser, args, output_manager, return_parser=False):
                         date = xtime = ''
                         item_name = subdir
                     if not options['totals']:
-                        output_manager.print_msg(
-                            "%s %10s %8s %24s %s",
-                            byte_str, date, xtime, content_type, item_name)
-                total_bytes += item_bytes
-
-        # report totals
-        if options['long'] or human:
-            if not container:
-                output_manager.print_msg(
-                    "%5s %s", prt_bytes(total_count, True),
-                    prt_bytes(total_bytes, human))
-            else:
-                output_manager.print_msg(
-                    prt_bytes(total_bytes, human))
+                        if options['versions']:
+                            output_manager.print_msg(
+                                "%s %10s %8s %16s %24s %s",
+                                byte_str, date, xtime,
+                                item.get('version_id', 'null'),
+                                content_type, item_name)
+                        else:
+                            output_manager.print_msg(
+                                "%s %10s %8s %24s %s",
+                                byte_str, date, xtime, content_type, item_name)
+                totals['bytes'] += item_bytes
 
     parser.add_argument(
         '-l', '--long', dest='long', action='store_true', default=False,
@@ -579,6 +601,8 @@ def st_list(parser, args, output_manager, return_parser=False):
              'what this means.')
     parser.add_argument('-j', '--json', action='store_true',
                         help='print listing information in json')
+    parser.add_argument('--versions', action='store_true',
+                        help='display all versions')
     parser.add_argument(
         '-H', '--header', action='append', dest='header',
         default=[],
@@ -592,6 +616,8 @@ def st_list(parser, args, output_manager, return_parser=False):
     args = args[1:]
     if options['delimiter'] and not args:
         exit('-d option only allowed for container listings')
+    if options['versions'] and not args:
+        exit('--versions option only allowed for container listings')
 
     human = options.pop('human')
     if human:
@@ -606,6 +632,7 @@ def st_list(parser, args, output_manager, return_parser=False):
         try:
             if not args:
                 stats_parts_gen = swift.list()
+                container = None
             else:
                 container = args[0]
                 args = args[1:]
@@ -631,17 +658,30 @@ def st_list(parser, args, output_manager, return_parser=False):
                     sort_keys=True, indent=2)
                 output_manager.print_msg('')
                 return
+
+            totals = {'count': 0, 'bytes': 0}
             for stats in stats_parts_gen:
                 if stats["success"]:
-                    _print_stats(options, stats, human)
+                    _print_stats(options, stats, human, totals)
                 else:
                     raise stats["error"]
 
+            # report totals
+            if options['long'] or human:
+                if container is None:
+                    output_manager.print_msg(
+                        "%12s %s", prt_bytes(totals['count'], True),
+                        prt_bytes(totals['bytes'], human))
+                else:
+                    output_manager.print_msg(
+                        prt_bytes(totals['bytes'], human))
+
         except SwiftError as e:
             output_manager.error(e.value)
 
 
 st_stat_options = '''[--lh] [--header <header:value>]
+                  [--version-id <version_id>]
                   [<container> [<object>]]
 '''
 
@@ -655,6 +695,8 @@ Positional arguments:
 Optional arguments:
   --lh                  Report sizes in human readable format similar to
                         ls -lh.
+  --version-id <version-id>
+                        Report stat of specific version of a versioned object.
   -H, --header <header:value>
                         Adds a custom request header to use for stat.
 '''.strip('\n')
@@ -664,6 +706,9 @@ def st_stat(parser, args, output_manager, return_parser=False):
     parser.add_argument(
         '--lh', dest='human', action='store_true', default=False,
         help='Report sizes in human readable format similar to ls -lh.')
+    parser.add_argument(
+        '--version-id', action='store', default=None,
+        help='Report stat of a specific version of a versioned object')
     parser.add_argument(
         '-H', '--header', action='append', dest='header',
         default=[],
@@ -675,6 +720,8 @@ def st_stat(parser, args, output_manager, return_parser=False):
 
     options, args = parse_args(parser, args)
     args = args[1:]
+    if options['version_id'] and len(args) < 2:
+        exit('--version-id option only allowed for object stats')
 
     with SwiftService(options=options) as swift:
         try:
@@ -1609,16 +1656,29 @@ def parse_args(parser, args, enforce_requires=True):
         return options, args
 
     if enforce_requires:
-        if options['auth_version'] == '3':
+        if options['os_auth_type'] and options['os_auth_type'] not in (
+                'password', 'v1password', 'v2password', 'v3password',
+                'v3applicationcredential'):
+            exit('Only "v3applicationcredential" is supported for '
+                 '--os-auth-type')
+        elif options['os_auth_type'] == 'v3applicationcredential':
+            if not (options['os_application_credential_id'] and
+                    options['os_application_credential_secret']):
+                exit('Auth version 3 (application credential) requires '
+                     'OS_APPLICATION_CREDENTIAL_ID and '
+                     'OS_APPLICATION_CREDENTIAL_SECRET to be set or '
+                     'overridden with --os-application-credential-id and '
+                     '--os-application-credential-secret respectively.')
+        elif options['auth_version'] == '3':
             if not options['auth']:
-                exit('Auth version 3 requires OS_AUTH_URL to be set or ' +
+                exit('Auth version 3 requires OS_AUTH_URL to be set or '
                      'overridden with --os-auth-url')
             if not (options['user'] or options['os_user_id']):
-                exit('Auth version 3 requires either OS_USERNAME or ' +
-                     'OS_USER_ID to be set or overridden with ' +
+                exit('Auth version 3 requires either OS_USERNAME or '
+                     'OS_USER_ID to be set or overridden with '
                      '--os-username or --os-user-id respectively.')
             if not options['key']:
-                exit('Auth version 3 requires OS_PASSWORD to be set or ' +
+                exit('Auth version 3 requires OS_PASSWORD to be set or '
                      'overridden with --os-password')
         elif not (options['auth'] and options['user'] and options['key']):
             exit('''
@@ -1789,6 +1849,29 @@ def add_default_args(parser):
                              'env[OS_AUTH_URL].')
     os_grp.add_argument('--os_auth_url',
                         help=argparse.SUPPRESS)
+    os_grp.add_argument('--os-auth-type',
+                        metavar='<auth-type>',
+                        default=environ.get('OS_AUTH_TYPE'),
+                        help='OpenStack auth type for v3. Defaults to '
+                             'env[OS_AUTH_TYPE].')
+    os_grp.add_argument('--os_auth_type',
+                        help=argparse.SUPPRESS)
+    os_grp.add_argument('--os-application-credential-id',
+                        metavar='<auth-application-credential-id>',
+                        default=environ.get('OS_APPLICATION_CREDENTIAL_ID'),
+                        help='OpenStack appplication credential id. '
+                             'Defaults to env[OS_APPLICATION_CREDENTIAL_ID].')
+    os_grp.add_argument('--os_application_credential_id',
+                        help=argparse.SUPPRESS)
+    os_grp.add_argument('--os-application-credential-secret',
+                        metavar='<auth-application-credential-secret>',
+                        default=environ.get(
+                            'OS_APPLICATION_CREDENTIAL_SECRET'),
+                        help='OpenStack appplication credential secret. '
+                             'Defaults to '
+                             'env[OS_APPLICATION_CREDENTIAL_SECRET].')
+    os_grp.add_argument('--os_application_credential_secret',
+                        help=argparse.SUPPRESS)
     os_grp.add_argument('--os-auth-token',
                         metavar='<auth-token>',
                         default=environ.get('OS_AUTH_TOKEN'),
@@ -1873,6 +1956,11 @@ def main(arguments=None):
              [--os-project-domain-name <auth-project-domain-name>]
              [--os-auth-url <auth-url>]
              [--os-auth-token <auth-token>]
+             [--os-auth-type <os-auth-type>]
+             [--os-application-credential-id
+                   <auth-application-credential-id>]
+             [--os-application-credential-secret
+                   <auth-application-credential-secret>]
              [--os-storage-url <storage-url>]
              [--os-region-name <region-name>]
              [--os-service-type <service-type>]
@@ -1925,6 +2013,11 @@ Examples:
       --os-user-id abcdef0123456789abcdef0123456789 \\
       --os-password password list
 
+  %(prog)s --os-auth-url https://api.example.com/v3 --auth-version 3\\
+      --os-application-credential-id d78683c92f0e4f9b9b02a2e208039412 \\
+      --os-application-credential-secret APPLICATION_CREDENTIAL_SECRET \\
+      --os-auth-type v3applicationcredential list
+
   %(prog)s --os-auth-token 6ee5eb33efad4e45ab46806eac010566 \\
       --os-storage-url https://10.1.5.2:8080/v1/AUTH_ced809b6a4baea7aeab61a \\
       list
@@ -1970,8 +2063,9 @@ Examples:
         try:
             globals()['st_%s' % args[0]](parser, argv[1:], output)
         except ClientException as err:
+            trans_id = err.transaction_id
+            err.transaction_id = None  # clear it so we aren't overly noisy
             output.error(str(err))
-            trans_id = (err.http_response_headers or {}).get('X-Trans-Id')
             if trans_id:
                 output.error("Failed Transaction ID: %s",
                              parse_header_string(trans_id))
